import gzip
from functools import partial

def flex_open(filename):
    """
    Attempts to determine whether a file is compressed, and 
    returns an open file object appropriate for the source 
    file type.

    Typical usage would be to replace calls to open() with 
    calls to flex_open(), e.g.

    with flex_open('compressed_file.gz') as f:
        {do stuff} 
    
    """
    magic_dict = {
        b'\x1f\x8b\x08': partial(gzip.open, mode='rt')
        # "\x42\x5a\x68": "bz2",
        # "\x50\x4b\x03\x04": "zip"
        }

    max_len = max(len(x) for x in magic_dict)
    
    open_func = None
    
    # Try opening the file in binary mode and reading the first
    # chunk to see if it matches the signature byte pattern
    # expected for each compressed file type

    with open(filename, 'rb') as f:
        file_start = f.read(max_len)
    for magic, func in magic_dict.items():
        if file_start.startswith(magic):
            open_func = func
    
    if open_func:
        return open_func(filename)
    else:
        return open(filename)


def is_fasta(some_file):
    with open(some_file) as f:
        last_was_header = False
        for l in f:
            if l.startswith("#"):
                continue
            if last_was_header is True:
                if not l.startswith('>'):
                    return True
                else:
                    return False
            elif l.startswith(">"):
                last_was_header = True
                continue
            else:
                return False


# authorship information
__author__ = 'Graham E. Larue'
__maintainer__ = "Graham E. Larue"
__email__ = 'egrahamlarue@gmail.com'
__license__ = 'GPL'

def translate(
    string, 
    verbosity="single",
    phase=0,
    mask=True,
    stop_char='*',
    as_codons=False,
    remove_whitespace=True,
    codon_map={
        'AAA': ('K', 'Lys', 'Lysine'),
        'AAC': ('N', 'Asn', 'Asparagine'),
        'AAG': ('K', 'Lys', 'Lysine'),
        'AAT': ('N', 'Asn', 'Asparagine'),
        'ACA': ('T', 'Thr', 'Threonine'),
        'ACC': ('T', 'Thr', 'Threonine'),
        'ACG': ('T', 'Thr', 'Threonine'),
        'ACT': ('T', 'Thr', 'Threonine'),
        'AGA': ('R', 'Arg', 'Arginine'),
        'AGC': ('S', 'Ser', 'Serine'),
        'AGG': ('R', 'Arg', 'Arginine'),
        'AGT': ('S', 'Ser', 'Serine'),
        'ATA': ('I', 'Ile', 'Isoleucine'),
        'ATC': ('I', 'Ile', 'Isoleucine'),
        'ATG': ('M', 'Met', 'Methionine'),
        'ATT': ('I', 'Ile', 'Isoleucine'),
        'CAA': ('Q', 'Gln', 'Glutamine'),
        'CAC': ('H', 'His', 'Histidine'),
        'CAG': ('Q', 'Gln', 'Glutamine'),
        'CAT': ('H', 'His', 'Histidine'),
        'CCA': ('P', 'Pro', 'Proline'),
        'CCC': ('P', 'Pro', 'Proline'),
        'CCG': ('P', 'Pro', 'Proline'),
        'CCT': ('P', 'Pro', 'Proline'),
        'CGA': ('R', 'Arg', 'Arginine'),
        'CGC': ('R', 'Arg', 'Arginine'),
        'CGG': ('R', 'Arg', 'Arginine'),
        'CGT': ('R', 'Arg', 'Arginine'),
        'CTA': ('L', 'Leu', 'Leucine'),
        'CTC': ('L', 'Leu', 'Leucine'),
        'CTG': ('L', 'Leu', 'Leucine'),
        'CTT': ('L', 'Leu', 'Leucine'),
        'GAA': ('E', 'Glu', 'Glutamate'),
        'GAC': ('D', 'Asp', 'Aspartate'),
        'GAG': ('E', 'Glu', 'Glutamate'),
        'GAT': ('D', 'Asp', 'Aspartate'),
        'GCA': ('A', 'Ala', 'Alanine'),
        'GCC': ('A', 'Ala', 'Alanine'),
        'GCG': ('A', 'Ala', 'Alanine'),
        'GCT': ('A', 'Ala', 'Alanine'),
        'GGA': ('G', 'Gly', 'Glycine'),
        'GGC': ('G', 'Gly', 'Glycine'),
        'GGG': ('G', 'Gly', 'Glycine'),
        'GGT': ('G', 'Gly', 'Glycine'),
        'GTA': ('V', 'Val', 'Valine'),
        'GTC': ('V', 'Val', 'Valine'),
        'GTG': ('V', 'Val', 'Valine'),
        'GTT': ('V', 'Val', 'Valine'),
        'TAC': ('Y', 'Tyr', 'Tyrosine'),
        'TAT': ('Y', 'Tyr', 'Tyrosine'),
        'TCA': ('S', 'Ser', 'Serine'),
        'TCC': ('S', 'Ser', 'Serine'),
        'TCG': ('S', 'Ser', 'Serine'),
        'TCT': ('S', 'Ser', 'Serine'),
        'TGC': ('C', 'Cys', 'Cysteine'),
        'TGG': ('W', 'Trp', 'Tryptophan'),
        'TGT': ('C', 'Cys', 'Cysteine'),
        'TTA': ('L', 'Leu', 'Leucine'),
        'TTC': ('F', 'Phe', 'Phenylalanine'),
        'TTG': ('L', 'Leu', 'Leucine'),
        'TTT': ('F', 'Phe', 'Phenylalanine')
    }):

    def _get_codons(s, p=0):
        for i in range(0, len(s), 3):
            codon = s[i + p:i + p + 3]
            if codon:  # don't return blank strings
                yield codon

    verbosity_index = {"single": 0, "short": 1, "long": 2}
    if remove_whitespace is True:
        string = ''.join(string.split())
    if as_codons:
        codons = _get_codons(string, p=phase)
        return ' '.join(codons)
    string = string.replace('U', 'T')
    codons = _get_codons(string, p=phase)
    stop_codons = ('TAG', 'TGA', 'TAA')
    amino_acids = []
    v = verbosity_index[verbosity]
    if v == 0:
        join_char = ""
    else:
        join_char = "-"
    for c in codons:
        aa = None
        try:
            aa = codon_map[c.upper()][v]
        except KeyError:
            if c.upper() in stop_codons:
                aa = stop_char
            elif mask:
                aa = 'X'
            else:
                aa = c.lower()
        amino_acids.append(aa)
            
    return join_char.join(amino_acids)


import re

def rev_comp(
    seq, 
    use_lower=True, 
    mask=None, 
    as_string=True, 
    mask_upper=re.compile('[^ATCG]'), 
    mask_lower=re.compile('[^ATCGatcg]')
):
    """
    Returns reverse complement of {seq}, with
    any non-ACTG characters replaced with {mask}

    If {as_string}, returns a string; else, returns 
    a list.

    """
    transform = {
        'A': 'T',
        'T': 'A',
        'C': 'G',
        'G': 'C',
        'N': 'N'
    }
    transform_mixed = {
        'A': 'T',
        'T': 'A',
        'C': 'G',
        'G': 'C',
        'N': 'N',
        'a': 't',
        't': 'a',
        'c': 'g',
        'g': 'c',
        'n': 'n'
    }
    if use_lower is True:
        mapper = transform_mixed
        masker = mask_lower
    else:
        mapper = transform
        masker = mask_upper
    if mask is not None:
        seq = re.sub(masker, mask, seq)
    mapper = str.maketrans(mapper)
    comp = seq.translate(mapper)
    reverse_comp = comp[::-1]
    if as_string is False:
        reverse_comp = list(reverse_comp)
    
    return reverse_comp


# import various functions here to allow for easy import
# from the biogl module (e.g. from biogl import {func})

from biogl.fasta_parse import fasta_parse
from biogl.GxfParse import GxfParse
from biogl.rev_comp import rev_comp
from biogl.translate import translate
from biogl.flex_open import flex_open
from biogl.random_seq import random_seq
from biogl.get_runtime import get_runtime
from biogl.byte_convert import byte_convert
from biogl.label_info import label_info
from biogl.is_fasta import is_fasta
from biogl.recursive_items import recursive_items
from biogl.window import window



def label_info(intron_label, *args):
    """
    Takes an intronIC label and any of the
    following keywords to specify desired
    information: "gene", "transcript",
    "species", "score", "index", "out_of".
    Returns a list of specified features.

    """
    import re
    
    reG=re.compile(r"^[^-]+-([^@]+)")
    reT=re.compile(r"^[^-]+-(?:[^@]+)@(.+)-intron")
    reP=re.compile(r"-intron_([^;]+)")
    species = intron_label.split("-")[0]
    try:
        score = intron_label.split(";")[-1].rstrip("%")
    except IndexError:
        score = None
    position = reP.findall(intron_label)[0]
    index = position.split("(")[0]
    out_of = position.split("(")[1].split(")")[0]
    gene = reG.search(intron_label).group(1)
    transcript = reT.search(intron_label).group(1)
    argMap = {  # associate variables with args
        "gene": gene,
        "transcript": transcript,
        "species": species,
        "score": score,
        "index": index,
        "out_of": out_of
    }
    intron_feats = []
    for arg in args:
        intron_feats.append(argMap[arg])

    return intron_feats